﻿using System;
using System.Collections.Generic;

using System.Text;

namespace Syncfusion.Windows.Forms.Diagram.Samples.DiagramTool
{
    public class DiagramEx: Syncfusion.Windows.Forms.Diagram.Controls.Diagram
    {
        public bool BringSelectionToFront(bool considerLayers)
        {
            //use instead of Controller.SendToBack
            if (considerLayers)
            {
                bool flag = false;
                if (!Controller.SelectionList.IsEmpty && (this.Model != null))
                {
                    Model.BeginUpdate();
                    Model model = this.Model;
                    model.HistoryManager.StartAtomicAction("BringToFront");
                    lock (Controller.SelectionList)
                    {
                        NodeCollection nodes = new NodeCollection();
                        nodes.AddRange(Controller.SelectionList);
                        nodes.Sort(new ReverseZOrderComparer());
                        foreach (Node node in nodes)
                        {
                            if (node.Layers.Count == 0)
                            {
                                model.BringToFront(node);
                            }
                            else
                            {
                                BringNodeToFrontOfLayer(node);
                            }
                            //model.SendToBack(node);
                        }
                        flag = true;
                    }

                    model.HistoryManager.EndAtomicAction();
                    EndUpdate();
                    Invalidate(true);
                }
                return flag;

            }
            else
            {
                return Controller.SendToBack();
            }
        }

        public bool BringSelectionForward(bool considerLayers)
        {
            if (considerLayers)
            {
                bool flag = false;
                if (!Controller.SelectionList.IsEmpty && (this.Model != null))
                {
                    Model model = this.Model;
                    model.BeginUpdate();
                    model.HistoryManager.StartAtomicAction("Bring Forward");
                    lock (Controller.SelectionList)
                    {
                        NodeCollection nodes = new NodeCollection();
                        nodes.AddRange(Controller.SelectionList);
                        nodes.Sort(new ReverseZOrderComparer());
                        foreach (Node node in nodes)
                        {
                            if (node.Layers.Count == 0)
                            {
                                model.BringForward(node);
                            }
                            else
                            {
                                BringNodeForwardInLayer(node);
                            }
                            //model.SendToBack(node);
                        }
                        flag = true;
                    }
                    model.HistoryManager.EndAtomicAction();
                    EndUpdate();
                    Invalidate(true);
                }
                return flag;

            }
            else
            {
                return Controller.BringForward();
            }
        }


        public bool SendSelectionBackward(bool considerLayers)
        {
            if (considerLayers)
            {
                bool flag = false;
                Model.BeginUpdate();
                if (!Controller.SelectionList.IsEmpty && (this.Model != null))
                {
                    Model model = this.Model;
                    model.HistoryManager.StartAtomicAction("Send Backward");
                    lock (Controller.SelectionList)
                    {
                        NodeCollection nodes = new NodeCollection();
                        nodes.AddRange(Controller.SelectionList);
                        nodes.Sort(new ZOrderComparer());
                        foreach (Node node in nodes)
                        {
                            if (node.Layers.Count == 0)
                            {
                                model.SendBackward(node);
                            }
                            else
                            {
                                SendNodeBackInLayer(node);
                            }
                            //model.SendToBack(node);
                        }
                        flag = true;
                    }
                    model.HistoryManager.EndAtomicAction();
                }
                EndUpdate();
                Invalidate(true);
                return flag;

            }
            else
            {
                return Controller.SendBackward();
            }
        }
        public void SendNodeToLayer(Layer destlayer, Node node)
        {
            Model.BeginUpdate();

            Layer currentActiveLayer = null;

            destlayer = Model.Layers.FindLayerByName(destlayer.Name);

            if (!destlayer.Enabled)
            {
                destlayer.Enabled = true;
                foreach (Layer layer in Model.Layers)
                {
                    if (destlayer != layer && layer.Enabled)
                    {
                        currentActiveLayer = layer;
                        layer.Enabled = false;
                    }
                }
            }
            MoveNodeToLayer_Internal(destlayer, node, null);

            if (currentActiveLayer != null)
            {
                destlayer.Enabled = false;
                currentActiveLayer.Enabled = true;
            }

            EndUpdate();
            Invalidate(true);
        }

        private void MoveNodeToLayer_Internal(Layer destlayer, Node node, int? newIndex)
        {
            if (!newIndex.HasValue)
            {
                newIndex = Model.Layers.IndexOf(destlayer);
            }

            if (node.Layers.Count == 0)
            {
                node.Layers.Add(destlayer);
            }
            else if (node.Layers[0].Name != destlayer.Name)
            {
                //determine if moving UP or down
                //if moving UP (i.e., layer 0 to layer 1)
                //then set zindex to lowest in new layer
                //if oving DOWN (layer 1 to 0)
                //then set zindex to highest in new layer

                Layer oldlayer = node.Layers[0];
                int oldindex = Model.Layers.IndexOf(oldlayer);

                oldlayer.Nodes.Remove(node);
                node.Layers.RemoveAt(0);

                node.Layers.Insert(0, destlayer);
                destlayer.Nodes.Add(node);

                if (oldindex > newIndex)
                {
                    //moving up, set zindex to lowest in new layer
                    SendNodeToBackOfLayer(node);
                }
                else if (oldindex < newIndex)
                {
                    BringNodeToFrontOfLayer(node);
                }
            }
        }

        public void SendSelectionToLayer(Layer destlayer)
        {
            Model.BeginUpdate();

            Layer currentActiveLayer = null;

            if (!destlayer.Enabled)
            {
                destlayer.Enabled = true;
                foreach (Layer layer in Model.Layers)
                {
                    if (destlayer != layer && layer.Enabled)
                    {
                        currentActiveLayer = layer;
                        layer.Enabled = false;
                    }
                }
            }

            int newIndex = Model.Layers.IndexOf(destlayer);
            foreach (Node node in Controller.SelectionList)
            {
                MoveNodeToLayer_Internal(destlayer, node, newIndex);
            }
            if (currentActiveLayer != null)
            {
                destlayer.Enabled = false;
                currentActiveLayer.Enabled = true;
            }

            EndUpdate();
            Invalidate(true);
        }

        public void MoveSelectionToBottomLayer()
        {
            if (Model.Layers.Count > 1)
            {
                Layer l = Model.Layers[0];
                SendSelectionToLayer(l);
            }
        }

        public void MoveSelectionUpOneLayer()
        {
            {

                if (Model.Layers.Count > 1)
                {
                    Model.HistoryManager.StartAtomicAction("Move UP One Layer");
                    foreach (Node n in Controller.SelectionList)
                    {
                        if (n.Layers.Count > 0)
                        {
                            Layer l = n.Layers[0];
                            l = Model.Layers.FindLayerByName(l.Name);   //get the real one

                            //determine the index
                            int indexof = Model.Layers.IndexOf(l);
                            if (indexof < Model.Layers.Count - 1)
                            {
                                SendNodeToLayer(Model.Layers[indexof + 1], n);
                            }
                        }
                        else
                        {
                            //put it on a layer if it's not already for some reason
                            if (Model.Layers.Count > 1)
                            {
                                SendSelectionToLayer(Model.Layers[1]);
                            }
                            else if (Model.Layers.Count > 0)
                            {
                                SendSelectionToLayer(Model.Layers[0]);
                            }
                        }
                    }

                    Model.HistoryManager.EndAtomicAction();
                }
            }

        }

        public void MoveSelectionDownOneLayer()
        {

            if (Model.Layers.Count > 1)
            {
                Model.HistoryManager.StartAtomicAction("Move Down One Layer");
                foreach (Node n in Controller.SelectionList)
                {
                    if (n.Layers.Count > 0)
                    {
                        Layer l = n.Layers[0];
                        l = Model.Layers.FindLayerByName(l.Name);   //get the real one

                        //determine the index
                        int indexof = Model.Layers.IndexOf(l);
                        if (indexof > 0)
                        {
                            SendNodeToLayer(Model.Layers[indexof - 1], n);
                        }
                    }
                }

                Model.HistoryManager.EndAtomicAction();
            }
        }
        private int BringNodeForwardInLayer(Node node)
        {
            if (node == null)
            {
                throw new ArgumentNullException("node");
            }
            int index = -1;

            if (node.Parent != Model)
            {
                IZOrderContainer parent = node.Parent as IZOrderContainer;
                if (parent != null)
                {
                    parent.BringForward(node);
                }
                return index;
            }
            //determine the lowest zorder in the node's current layer
            Layer layer = node.Layers[0];
            layer = Model.Layers[layer.Name];
            int maxZ = node.ZOrder;
            for (int i = 0; i < layer.Nodes.Count; i++)
            {
                if (layer.Nodes[i] != node && layer.Nodes[i].ZOrder > maxZ)
                {
                    maxZ = layer.Nodes[i].ZOrder;
                }
            }
            if (maxZ > node.ZOrder)
            {
                index = Model.BringForward(node);
            }
            return index;
        }

        private int SendNodeBackInLayer(Node node)
        {
            if (node == null)
            {
                throw new ArgumentNullException("node");
            }
            int index = -1;

            if (node.Parent != Model)
            {
                IZOrderContainer parent = node.Parent as IZOrderContainer;
                if (parent != null)
                {
                    parent.SendBackward(node);
                }
                return index;
            }
            //determine the lowest zorder in the node's current layer
            Layer layer = node.Layers[0];
            layer = Model.Layers[layer.Name];
            int maxZ = node.ZOrder;
            for (int i = 0; i < layer.Nodes.Count; i++)
            {
                if (layer.Nodes[i] != node && layer.Nodes[i].ZOrder < maxZ)
                {
                    maxZ = layer.Nodes[i].ZOrder;
                }
            }
            if (maxZ < node.ZOrder)
            {
                index = Model.SendBackward(node);
            }
            return index;
        }


        public bool SendSelectionToBack(bool considerLayers)
        {
            //use instead of Controller.SendToBack
            if (considerLayers)
            {
                Model.BeginUpdate();
                bool flag = false;
                if (!Controller.SelectionList.IsEmpty && (this.Model != null))
                {
                    Model model = this.Model;
                    model.HistoryManager.StartAtomicAction("SendToBack");
                    lock (Controller.SelectionList)
                    {
                        NodeCollection nodes = new NodeCollection();
                        nodes.AddRange(Controller.SelectionList);
                        nodes.Sort(new ZOrderComparer());
                        foreach (Node node in nodes)
                        {
                            if (node.Layers.Count == 0)
                            {
                                model.SendToBack(node);
                            }
                            else
                            {
                                SendNodeToBackOfLayer(node);
                            }
                            //model.SendToBack(node);
                        }
                        flag = true;
                    }
                    model.HistoryManager.EndAtomicAction();
                }

                EndUpdate();
                Invalidate(true);
                return flag;

            }
            else
            {
                return Controller.SendToBack();
            }
        }

        /// <summary>
        /// only concerned with the node's First Layer (Layers[0])
        /// Determines lowest Zindex of items in this layer
        /// sends node to that zorder
        /// </summary>
        /// <param name="node"></param>
        /// <returns></returns>
        public int SendNodeToBackOfLayer(Node node)
        {
            if (node == null)
            {
                throw new ArgumentNullException("node");
            }
            int index = -1;

            if (node.Parent != Model)
            {
                IZOrderContainer parent = node.Parent as IZOrderContainer;
                if (parent != null)
                {
                    parent.SendToBack(node);
                }
                return index;
            }
            //determine the lowest zorder in the node's current layer
            Layer layer = node.Layers[0];
            layer = Model.Layers[layer.Name];
            int minZindex = node.ZOrder;

            for (int i = 0; i < layer.Nodes.Count; i++)
            {
                if (layer.Nodes[i] != node && layer.Nodes[i].ZOrder < minZindex)
                {
                    minZindex = layer.Nodes[i].ZOrder;
                }
            }
            //must also determine minindex in layer above this one
            int layerindex = Model.Layers.IndexOf(layer);
            if (layerindex < Model.Layers.Count - 1)
            {
                Layer nextLayer = Model.Layers[layerindex + 1];
                for (int i = 0; i < nextLayer.Nodes.Count; i++)
                {
                    if (nextLayer.Nodes[i] != node && nextLayer.Nodes[i].ZOrder < minZindex)
                    {
                        minZindex = nextLayer.Nodes[i].ZOrder;
                    }
                }
            }

            if (minZindex < node.ZOrder)
            {
                index = Model.SetZOrder(node, minZindex);
            }

            return index;
        }

        /// <summary>
        /// only concerned with the node's First Layer (Layers[0])
        /// Determines max Zindex of items in this layer
        /// sends node to that zorder
        /// </summary>
        /// <param name="node"></param>
        /// <returns></returns>
        public int BringNodeToFrontOfLayer(Node node)
        {
            if (node == null)
            {
                throw new ArgumentNullException("node");
            }
            int index = -1;

            if (node.Parent != Model)
            {
                IZOrderContainer parent = node.Parent as IZOrderContainer;
                if (parent != null)
                {
                    parent.BringToFront(node);
                }
                return index;
            }
            //determine the lowest zorder in the node's current layer
            Layer layer = node.Layers[0];
            layer = Model.Layers[layer.Name];
            int maxZ = node.ZOrder;
            for (int i = 0; i < layer.Nodes.Count; i++)
            {
                if (layer.Nodes[i] != node && layer.Nodes[i].ZOrder > maxZ)
                {
                    maxZ = layer.Nodes[i].ZOrder;
                }
            }

            //now determine for next lowest layer just in case
            int layerIndex = Model.Layers.IndexOf(layer);
            if (layerIndex > 0)
            {
                Layer nextLayer = Model.Layers[layerIndex - 1];
                for (int i = 0; i < nextLayer.Nodes.Count; i++)
                {
                    if (nextLayer.Nodes[i] != node && nextLayer.Nodes[i].ZOrder > maxZ)
                    {
                        maxZ = nextLayer.Nodes[i].ZOrder;
                    }
                }
            }
            if (maxZ > node.ZOrder)
            {
                index = Model.SetZOrder(node, maxZ);
            }
            return index;
        }
    }
}
